# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pytest

from mockito.invocation import InvocationError
from mockito import mock, when, verify


class Foo(object):
    def bar(self):
        pass

class Action(object):
    def no_arg(self):
        pass

    def run(self, arg):
        return arg

    def __call__(self, task):
        return task


class TestSpeccing:
    def testStubCallAndVerify(self):
        action = mock(Action)

        when(action).run(11).thenReturn(12)
        assert action.run(11) == 12
        verify(action).run(11)


    def testShouldScreamWhenStubbingUnknownMethod(self):
        action = mock(Action)

        with pytest.raises(InvocationError):
            when(action).unknownMethod()

    def testShouldScreamWhenCallingUnknownMethod(self):
        action = mock(Action)

        with pytest.raises(AttributeError):
            action.unknownMethod()

    def testShouldScreamWhenCallingUnexpectedMethod(self):
        action = mock(Action)

        with pytest.raises(AttributeError):
            action.run(11)

    def testPreconfigureMockWithAttributes(self):
        action = mock({'foo': 'bar'}, spec=Action)

        assert action.foo == 'bar'
        with pytest.raises(InvocationError):
            when(action).remember()

    def testPreconfigureWithFunction(self):
        action = mock({
            'run': lambda _: 12
        }, spec=Action)

        assert action.run(11) == 12

        verify(action).run(11)

    def testPreconfigureWithFunctionThatTakesNoArgs(self):
        action = mock({
            'no_arg': lambda: 12
        }, spec=Action)

        assert action.no_arg() == 12

        verify(action).no_arg()

    def testShouldScreamOnUnknownAttribute(self):
        action = mock(Action)

        with pytest.raises(AttributeError):
            action.cam

    def testShouldPassIsInstanceChecks(self):
        action = mock(Action)

        assert isinstance(action, Action)

    def testShouldPassIsInstanceChecks_2(self):
        action = mock(spec=Action)

        assert isinstance(action, Action)

    def testShouldPassIsInstanceChecks_3(self):
        action = mock({}, Action)

        assert isinstance(action, Action)

    def testShouldPassIsInstanceChecks_4(self):
        action = mock({}, spec=Action)

        assert isinstance(action, Action)

    def testHasANiceName(self):
        action = mock(Action)

        assert repr(action) == "<DummyAction id=%s>" % id(action)


class TestSpeccingLoose:
    def testReturnNoneForEveryMethod(self):
        action = mock(Action, strict=False)
        assert action.unknownMethod() is None
        assert action.run(11) is None





import pytest

from mockito.utils import get_obj, get_obj_attr_tuple

import sys


PY3 = sys.version_info >= (3,)


def foo():
    pass

class TestLateImports:

    def testOs(self):
        import os
        assert get_obj('os') is os

    def testOsPath(self):
        import os.path
        assert get_obj('os.path') is os.path

    def testOsPathExists(self):
        import os.path
        assert get_obj('os.path.exists') is os.path.exists

    def testOsPathWhatever(self):
        with pytest.raises(AttributeError) as exc:
            get_obj('os.path.whatever')

        assert str(exc.value) == "module 'os.path' has no attribute 'whatever'"

    def testOsPathExistsForever(self):
        with pytest.raises(AttributeError) as exc:
            get_obj('os.path.exists.forever')

        assert str(exc.value) == \
            "object 'os.path.exists' has no attribute 'forever'"

    def testOsPathExistsForeverAndEver(self):
        with pytest.raises(AttributeError) as exc:
            get_obj('os.path.exists.forever.and.ever')

        assert str(exc.value) == \
            "object 'os.path.exists' has no attribute 'forever'"

    def testUnknownMum(self):
        with pytest.raises(ImportError) as exc:
            assert get_obj('mum') is foo
        if PY3:
            assert str(exc.value) == "No module named 'mum'"
        else:
            assert str(exc.value) == "No module named mum"

    def testUnknownMumFoo(self):
        with pytest.raises(ImportError) as exc:
            assert get_obj('mum.foo') is foo
        if PY3:
            assert str(exc.value) == "No module named 'mum'"
        else:
            assert str(exc.value) == "No module named mum"

    def testReturnGivenObject(self):
        import os
        assert get_obj(os) == os
        assert get_obj(os.path) == os.path
        assert get_obj(2) == 2

    def testDisallowRelativeImports(self):
        with pytest.raises(TypeError):
            get_obj('.mum')

class TestReturnTuple:
    def testOs(self):
        with pytest.raises(TypeError):
            get_obj_attr_tuple('os')

    def testOsPath(self):
        import os
        assert get_obj_attr_tuple('os.path') == (os, 'path')

    def testOsPathExists(self):
        import os
        assert get_obj_attr_tuple('os.path.exists') == (os.path, 'exists')

    def testOsPathExistsNot(self):
        import os
        assert get_obj_attr_tuple('os.path.exists.not') == (
            os.path.exists, 'not')

    def testDisallowRelativeImports(self):
        with pytest.raises(TypeError):
            get_obj('.mum')





# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pytest
from mockito import any, mock, times, verify, when

from .test_base import TestBase


class TestEmptyMocks:
    def testAllMethodsReturnNone(self):
        dummy = mock()

        assert dummy.foo() is None
        assert dummy.foo(1, 2) is None


    def testConfigureDummy(self):
        dummy = mock({'foo': 'bar'})
        assert dummy.foo == 'bar'

    def testConfigureDummyWithFunction(self):
        dummy = mock({
            'getStuff': lambda s: s + ' thing'
        })

        assert dummy.getStuff('da') == 'da thing'
        verify(dummy).getStuff('da')

    def testDummiesAreCallable(self):
        dummy = mock()
        assert dummy() is None
        assert dummy(1, 2) is None

    def testCallsAreVerifiable(self):
        dummy = mock()
        dummy(1, 2)

        verify(dummy).__call__(1, 2)

    def testConfigureCallBehavior(self):
        dummy = mock()
        when(dummy).__call__(1).thenReturn(2)

        assert dummy(1) == 2
        verify(dummy).__call__(1)

    def testCheckIsInstanceAgainstItself(self):
        dummy = mock()
        assert isinstance(dummy, dummy.__class__)


    def testConfigureMagicMethod(self):
        dummy = mock()
        when(dummy).__getitem__(1).thenReturn(2)

        assert dummy[1] == 2

class TestStrictEmptyMocks:
    def testScream(self):
        dummy = mock(strict=True)

        with pytest.raises(AttributeError):
            dummy.foo()

    def testAllowStubbing(self):
        dummy = mock(strict=True)
        when(dummy).foo()
        dummy.foo()
        verify(dummy).foo()

    def testCanConfigureCall(self):
        dummy = mock(strict=True)
        when(dummy).__call__(1).thenReturn(2)

        assert dummy(1) == 2

    def testScreamOnUnconfiguredCall(self):
        dummy = mock(strict=True)

        with pytest.raises(AttributeError):
            dummy(1)

    def testConfigureMagicMethod(self):
        dummy = mock(strict=True)
        when(dummy).__getitem__(1).thenReturn(2)

        assert dummy[1] == 2


class StubbingTest(TestBase):
    def testStubsWithReturnValue(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo")
        when(theMock).getMoreStuff(1, 2).thenReturn(10)

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual(10, theMock.getMoreStuff(1, 2))
        self.assertEqual(None, theMock.getMoreStuff(1, 3))

    def testStubsWhenNoArgsGiven(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo")
        when(theMock).getWidget().thenReturn("bar")

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual("bar", theMock.getWidget())

    def testStubsConsecutivelyWhenNoArgsGiven(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo").thenReturn("bar")
        when(theMock).getWidget().thenReturn("baz").thenReturn("baz2")

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual("bar", theMock.getStuff())
        self.assertEqual("bar", theMock.getStuff())
        self.assertEqual("baz", theMock.getWidget())
        self.assertEqual("baz2", theMock.getWidget())
        self.assertEqual("baz2", theMock.getWidget())

    def testStubsWithException(self):
        theMock = mock()
        when(theMock).someMethod().thenRaise(Exception("foo"))

        self.assertRaisesMessage("foo", theMock.someMethod)

    def testStubsAndVerifies(self):
        theMock = mock()
        when(theMock).foo().thenReturn("foo")

        self.assertEqual("foo", theMock.foo())
        verify(theMock).foo()

    def testStubsVerifiesAndStubsAgain(self):
        theMock = mock()

        when(theMock).foo().thenReturn("foo")
        self.assertEqual("foo", theMock.foo())
        verify(theMock).foo()

        when(theMock).foo().thenReturn("next foo")
        self.assertEqual("next foo", theMock.foo())
        verify(theMock, times(2)).foo()

    def testOverridesStubbing(self):
        theMock = mock()

        when(theMock).foo().thenReturn("foo")
        when(theMock).foo().thenReturn("bar")

        self.assertEqual("bar", theMock.foo())

    def testStubsAndInvokesTwiceAndVerifies(self):
        theMock = mock()

        when(theMock).foo().thenReturn("foo")

        self.assertEqual("foo", theMock.foo())
        self.assertEqual("foo", theMock.foo())

        verify(theMock, times(2)).foo()

    def testStubsAndReturnValuesForSameMethodWithDifferentArguments(self):
        theMock = mock()
        when(theMock).getStuff(1).thenReturn("foo")
        when(theMock).getStuff(1, 2).thenReturn("bar")

        self.assertEqual("foo", theMock.getStuff(1))
        self.assertEqual("bar", theMock.getStuff(1, 2))

    def testStubsAndReturnValuesForSameMethodWithDifferentNamedArguments(self):
        repo = mock()
        when(repo).findby(id=6).thenReturn("John May")
        when(repo).findby(name="John").thenReturn(["John May", "John Smith"])

        self.assertEqual("John May", repo.findby(id=6))
        self.assertEqual(["John May", "John Smith"], repo.findby(name="John"))

    def testStubsForMethodWithSameNameAndNamedArgumentsInArbitraryOrder(self):
        theMock = mock()

        when(theMock).foo(first=1, second=2, third=3).thenReturn(True)

        self.assertEqual(True, theMock.foo(third=3, first=1, second=2))

    def testStubsMethodWithSameNameAndMixedArguments(self):
        repo = mock()
        when(repo).findby(1).thenReturn("John May")
        when(repo).findby(1, active_only=True).thenReturn(None)
        when(repo).findby(name="Sarah").thenReturn(["Sarah Connor"])
        when(repo).findby(name="Sarah", active_only=True).thenReturn([])

        self.assertEqual("John May", repo.findby(1))
        self.assertEqual(None, repo.findby(1, active_only=True))
        self.assertEqual(["Sarah Connor"], repo.findby(name="Sarah"))
        self.assertEqual([], repo.findby(name="Sarah", active_only=True))

    def testStubsWithChainedReturnValues(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo") \
                                .thenReturn("bar") \
                                .thenReturn("foobar")

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual("bar", theMock.getStuff())
        self.assertEqual("foobar", theMock.getStuff())

    def testStubsWithChainedReturnValuesAndException(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo") \
                                .thenReturn("bar") \
                                .thenRaise(Exception("foobar"))

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual("bar", theMock.getStuff())
        self.assertRaisesMessage("foobar", theMock.getStuff)

    def testStubsWithChainedExceptionAndReturnValue(self):
        theMock = mock()
        when(theMock).getStuff().thenRaise(Exception("foo")) \
                                .thenReturn("bar")

        self.assertRaisesMessage("foo", theMock.getStuff)
        self.assertEqual("bar", theMock.getStuff())

    def testStubsWithChainedExceptions(self):
        theMock = mock()
        when(theMock).getStuff().thenRaise(Exception("foo")) \
                                .thenRaise(Exception("bar"))

        self.assertRaisesMessage("foo", theMock.getStuff)
        self.assertRaisesMessage("bar", theMock.getStuff)

    def testStubsWithReturnValueBeingException(self):
        theMock = mock()
        exception = Exception("foo")
        when(theMock).getStuff().thenReturn(exception)

        self.assertEqual(exception, theMock.getStuff())

    def testLastStubbingWins(self):
        theMock = mock()
        when(theMock).foo().thenReturn(1)
        when(theMock).foo().thenReturn(2)

        self.assertEqual(2, theMock.foo())

    def testStubbingOverrides(self):
        theMock = mock()
        when(theMock).foo().thenReturn(1)
        when(theMock).foo().thenReturn(2).thenReturn(3)

        self.assertEqual(2, theMock.foo())
        self.assertEqual(3, theMock.foo())
        self.assertEqual(3, theMock.foo())

    def testStubsWithMatchers(self):
        theMock = mock()
        when(theMock).foo(any()).thenReturn(1)

        self.assertEqual(1, theMock.foo(1))
        self.assertEqual(1, theMock.foo(100))

    def testStubbingOverrides2(self):
        theMock = mock()
        when(theMock).foo(any()).thenReturn(1)
        when(theMock).foo("oh").thenReturn(2)

        self.assertEqual(2, theMock.foo("oh"))
        self.assertEqual(1, theMock.foo("xxx"))

    def testDoesNotVerifyStubbedCalls(self):
        theMock = mock()
        when(theMock).foo().thenReturn(1)

        verify(theMock, times=0).foo()

    def testStubsWithMultipleReturnValues(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo", "bar", "foobar")

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual("bar", theMock.getStuff())
        self.assertEqual("foobar", theMock.getStuff())

    def testStubsWithChainedMultipleReturnValues(self):
        theMock = mock()
        when(theMock).getStuff().thenReturn("foo", "bar") \
                                .thenReturn("foobar")

        self.assertEqual("foo", theMock.getStuff())
        self.assertEqual("bar", theMock.getStuff())
        self.assertEqual("foobar", theMock.getStuff())

    def testStubsWithMultipleExceptions(self):
        theMock = mock()
        when(theMock).getStuff().thenRaise(Exception("foo"), Exception("bar"))

        self.assertRaisesMessage("foo", theMock.getStuff)
        self.assertRaisesMessage("bar", theMock.getStuff)

    def testStubsWithMultipleChainedExceptions(self):
        theMock = mock()
        when(theMock).getStuff() \
                     .thenRaise(Exception("foo"), Exception("bar")) \
                     .thenRaise(Exception("foobar"))

        self.assertRaisesMessage("foo", theMock.getStuff)
        self.assertRaisesMessage("bar", theMock.getStuff)
        self.assertRaisesMessage("foobar", theMock.getStuff)

    def testLeavesOriginalMethodUntouchedWhenCreatingStubFromRealClass(self):
        class Person:
            def get_name(self):
                return "original name"

        # given
        person = Person()
        mockPerson = mock(Person)

        # when
        when(mockPerson).get_name().thenReturn("stubbed name")

        # then
        self.assertEqual("stubbed name", mockPerson.get_name())
        self.assertEqual("original name", person.get_name(),
                         'Original method should not be replaced.')

    def testStubsWithThenAnswer(self):
        m = mock()

        when(m).magic_number().thenAnswer(lambda: 5)

        self.assertEqual(m.magic_number(), 5)

        when(m).add_one(any()).thenAnswer(lambda number: number + 1)

        self.assertEqual(m.add_one(5), 6)
        self.assertEqual(m.add_one(8), 9)

        when(m).do_times(any(), any()).thenAnswer(lambda one, two: one * two)

        self.assertEqual(m.do_times(5, 4), 20)
        self.assertEqual(m.do_times(8, 5), 40)

        when(m).do_dev_magic(any(), any()).thenAnswer(lambda a, b: a / b)

        self.assertEqual(m.do_dev_magic(20, 4), 5)
        self.assertEqual(m.do_dev_magic(40, 5), 8)

        def test_key_words(testing="Magic"):
            return testing + " Stuff"

        when(m).with_key_words().thenAnswer(test_key_words)
        self.assertEqual(m.with_key_words(), "Magic Stuff")

        when(m).with_key_words(testing=any()).thenAnswer(test_key_words)
        self.assertEqual(m.with_key_words(testing="Very Funky"),
                         "Very Funky Stuff")

    def testStubsWithThenAnswerAndMixedArgs(self):
        repo = mock()

        def method_one(value, active_only=False):
            return None

        def method_two(name=None, active_only=False):
            return ["%s Connor" % name]

        def method_three(name=None, active_only=False):
            return [name, active_only, 0]

        when(repo).findby(1).thenAnswer(lambda x: "John May (%d)" % x)
        when(repo).findby(1, active_only=True).thenAnswer(method_one)
        when(repo).findby(name="Sarah").thenAnswer(method_two)
        when(repo).findby(
            name="Sarah", active_only=True).thenAnswer(method_three)

        self.assertEqual("John May (1)", repo.findby(1))
        self.assertEqual(None, repo.findby(1, active_only=True))
        self.assertEqual(["Sarah Connor"], repo.findby(name="Sarah"))
        self.assertEqual(
            ["Sarah", True, 0], repo.findby(name="Sarah", active_only=True))




# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from mockito.matchers import MatcherError
from .test_base import TestBase
from mockito import mock, verify, when
from mockito.matchers import and_, or_, not_, eq, neq, lt, lte, gt, gte, \
    any_, arg_that, contains, matches, captor, ANY, ARGS, KWARGS
import re


class TestConvenienceMatchers:
    def testBuiltinAnyStandsForOurAny(self):
        dummy = mock()
        dummy.foo(1)
        dummy.foo('a')
        dummy.foo(True)

        verify(dummy, times=3).foo(any)

        dummy.foo(a=12)
        verify(dummy).foo(a=any)

    def testOurAnyCanBeUsedAsAType(self):
        dummy = mock()
        dummy.foo(1)
        dummy.foo('a')
        dummy.foo(True)
        verify(dummy, times=3).foo(any_)


class TestAliases:
    def testANY(self):
        dummy = mock()
        dummy.foo(1)
        verify(dummy).foo(ANY)

    def testARGS(self):
        dummy = mock()
        dummy.foo(1)
        verify(dummy).foo(*ARGS)

    def testKWARGS(self):
        dummy = mock()
        dummy.foo(a=1)
        verify(dummy).foo(**KWARGS)


class MatchersTest(TestBase):
    def testVerifiesUsingContainsMatcher(self):
        ourMock = mock()
        ourMock.foo("foobar")

        verify(ourMock).foo(contains("foo"))
        verify(ourMock).foo(contains("bar"))


class AndMatcherTest(TestBase):
    def testShouldSatisfyIfAllMatchersAreSatisfied(self):
        self.assertTrue(
            and_(contains("foo"), contains("bar")).matches("foobar"))

    def testShouldNotSatisfyIfOneOfMatchersIsNotSatisfied(self):
        self.assertFalse(
            and_(contains("foo"), contains("bam")).matches("foobar"))

    def testShouldTreatNonMatchersAsEqMatcher(self):
        self.assertTrue(and_("foo", any_(str)).matches("foo"))
        self.assertFalse(and_("foo", any_(int)).matches("foo"))


class OrMatcherTest(TestBase):
    def testShouldSatisfyIfAnyOfMatchersIsSatisfied(self):
        self.assertTrue(
            or_(contains("foo"), contains("bam")).matches("foobar"))

    def testShouldNotSatisfyIfAllOfMatchersAreNotSatisfied(self):
        self.assertFalse(
            or_(contains("bam"), contains("baz")).matches("foobar"))

    def testShouldTreatNonMatchersAsEqMatcher(self):
        self.assertTrue(or_("foo", "bar").matches("foo"))
        self.assertFalse(or_("foo", "bar").matches("bam"))


class NotMatcherTest(TestBase):
    def testShouldSatisfyIfInnerMatcherIsNotSatisfied(self):
        self.assertTrue(not_(contains("foo")).matches("bar"))

    def testShouldNotSatisfyIfInnerMatcherIsSatisfied(self):
        self.assertFalse(not_(contains("foo")).matches("foo"))

    def testShouldTreatNonMatchersAsEqMatcher(self):
        self.assertTrue(or_("foo", "bar").matches("foo"))


class EqMatcherTest(TestBase):
    def testShouldSatisfyIfArgMatchesGivenValue(self):
        self.assertTrue(eq("foo").matches("foo"))

    def testShouldNotSatisfyIfArgDoesNotMatchGivenValue(self):
        self.assertFalse(eq("foo").matches("bar"))


class NeqMatcherTest(TestBase):
    def testShouldSatisfyIfArgDoesNotMatchGivenValue(self):
        self.assertTrue(neq("foo").matches("bar"))

    def testShouldNotSatisfyIfArgMatchesGivenValue(self):
        self.assertFalse(neq("foo").matches("foo"))


class LtMatcherTest(TestBase):
    def testShouldSatisfyIfArgIsLessThanGivenValue(self):
        self.assertTrue(lt(5).matches(4))

    def testShouldNotSatisfyIfArgIsEqualToGivenValue(self):
        self.assertFalse(lt(5).matches(5))

    def testShouldNotSatisfyIfArgIsGreaterThanGivenValue(self):
        self.assertFalse(lt(5).matches(6))


class LteMatcherTest(TestBase):
    def testShouldSatisfyIfArgIsLessThanGivenValue(self):
        self.assertTrue(lte(5).matches(4))

    def testShouldSatisfyIfArgIsEqualToGivenValue(self):
        self.assertTrue(lte(5).matches(5))

    def testShouldNotSatisfyIfArgIsGreaterThanGivenValue(self):
        self.assertFalse(lte(5).matches(6))


class GtMatcherTest(TestBase):
    def testShouldNotSatisfyIfArgIsLessThanGivenValue(self):
        self.assertFalse(gt(5).matches(4))

    def testShouldNotSatisfyIfArgIsEqualToGivenValue(self):
        self.assertFalse(gt(5).matches(5))

    def testShouldSatisfyIfArgIsGreaterThanGivenValue(self):
        self.assertTrue(gt(5).matches(6))


class GteMatcherTest(TestBase):
    def testShouldNotSatisfyIfArgIsLessThanGivenValue(self):
        self.assertFalse(gte(5).matches(4))

    def testShouldSatisfyIfArgIsEqualToGivenValue(self):
        self.assertTrue(gte(5).matches(5))

    def testShouldSatisfyIfArgIsGreaterThanGivenValue(self):
        self.assertTrue(gte(5).matches(6))


class ArgThatMatcherTest(TestBase):
    def testShouldSatisfyIfPredicateReturnsTrue(self):
        self.assertTrue(arg_that(lambda arg: arg > 5).matches(10))

    def testShouldNotSatisfyIfPredicateReturnsFalse(self):
        self.assertFalse(arg_that(lambda arg: arg > 5).matches(1))


class ContainsMatcherTest(TestBase):
    def testShouldSatisfiySubstringOfGivenString(self):
        self.assertTrue(contains("foo").matches("foobar"))

    def testShouldSatisfySameString(self):
        self.assertTrue(contains("foobar").matches("foobar"))

    def testShouldNotSatisfiyStringWhichIsNotSubstringOfGivenString(self):
        self.assertFalse(contains("barfoo").matches("foobar"))

    def testShouldNotSatisfiyEmptyString(self):
        self.assertFalse(contains("").matches("foobar"))

    def testShouldNotSatisfiyNone(self):
        self.assertFalse(contains(None).matches("foobar"))


class MatchesMatcherTest(TestBase):
    def testShouldSatisfyIfRegexMatchesGivenString(self):
        self.assertTrue(matches('f..').matches('foo'))

    def testShouldAllowSpecifyingRegexFlags(self):
        self.assertFalse(matches('f..').matches('Foo'))
        self.assertTrue(matches('f..', re.IGNORECASE).matches('Foo'))

    def testShouldNotSatisfyIfRegexIsNotMatchedByGivenString(self):
        self.assertFalse(matches('f..').matches('bar'))


class ArgumentCaptorTest(TestBase):
    def test_matches_anything_by_default(self):
        assert captor().matches(12)
        assert captor().matches("anything")
        assert captor().matches(int)

    def test_matches_is_constrained_by_inner_matcher(self):
        assert captor(any_(int)).matches(12)
        assert not captor(any_(int)).matches("12")

    def test_all_values_initially_is_empty(self):
        c = captor()
        assert c.all_values == []

    def test_captures_all_values(self):
        m = mock()
        c = captor()

        when(m).do(c)
        m.do("any")
        m.do("thing")

        assert c.all_values == ["any", "thing"]

    def test_captures_only_matching_values(self):
        m = mock()
        c = captor(any_(int))

        when(m).do(c)
        m.do("any")
        m.do("thing")
        m.do(21)

        assert c.all_values == [21]

    def test_captures_all_values_while_verifying(self):
        m = mock()
        c = captor()

        m.do("any")
        m.do("thing")
        verify(m, times=2).do(c)

        assert c.all_values == ["any", "thing"]

    def test_remember_last_value(self):
        m = mock()
        c = captor()

        when(m).do(c)
        m.do("any")
        m.do("thing")

        assert c.value == "thing"

    def test_remember_last_value_while_verifying(self):
        m = mock()
        c = captor()

        m.do("any")
        m.do("thing")
        verify(m, times=2).do(c)

        assert c.value == "thing"

    def test_accessing_value_throws_if_nothing_captured_yet(self):
        c = captor()
        with self.assertRaises(MatcherError):
            _ = c.value

    def test_expose_issue_49_using_when(self):
        m = mock()
        c = captor()

        when(m).do(c, 10)
        when(m).do(c, 11)
        m.do("anything", 10)

        assert c.all_values == ["anything"]

    def test_expose_issue_49_using_verify(self):
        m = mock()
        c = captor()

        m.do("anything", 10)
        verify(m).do(c, 10)
        verify(m, times=0).do(c, 11)

        assert c.all_values == ["anything"]




from collections import namedtuple

import pytest
from mockito import args, invocation, kwargs, mock, when


class Dog(object):
    def bark(self, sound):
        return "%s!" % sound

    def waggle(self):
        return 'waggle'

class CallSignature(namedtuple('CallSignature', 'args kwargs')):
    def raises(self, reason):
        return pytest.mark.xfail(self, raises=reason, strict=True)

def sig(*args, **kwargs):
    return CallSignature(args, kwargs)


class TestCallMethodWithSignature:
    def testNoArg(self):
        rex = Dog()
        when(rex).waggle().thenReturn('wuff')

        assert rex.waggle() == 'wuff'

        with pytest.raises(TypeError):
            rex.waggle(1)
        with pytest.raises(TypeError):
            rex.waggle(Ellipsis)
        with pytest.raises(TypeError):
            rex.waggle(args)
        with pytest.raises(TypeError):
            rex.waggle(kwargs)
        with pytest.raises(TypeError):
            rex.waggle(*args)
        with pytest.raises(TypeError):
            rex.waggle(**kwargs)

    def testExpectingSpecificInputAsPositionalArgument(self):
        rex = Dog()
        when(rex).bark(1).thenReturn('wuff')

        assert rex.bark(1) == 'wuff'

        with pytest.raises(invocation.InvocationError):
            rex.bark(sound=1)
        with pytest.raises(invocation.InvocationError):
            rex.bark(Ellipsis)
        with pytest.raises(invocation.InvocationError):
            rex.bark(args)
        with pytest.raises(invocation.InvocationError):
            rex.bark(*args)
        with pytest.raises(invocation.InvocationError):
            rex.bark(kwargs)

        with pytest.raises(TypeError):
            rex.bark(1, 2)
        with pytest.raises(TypeError):
            rex.bark(wuff=1)
        with pytest.raises(TypeError):
            rex.bark(**kwargs)

    def testExpectingSpecificInputAsKeyword(self):
        rex = Dog()
        when(rex).bark(sound=1).thenReturn('wuff')

        assert rex.bark(sound=1) == 'wuff'

        with pytest.raises(invocation.InvocationError):
            rex.bark(1)
        with pytest.raises(invocation.InvocationError):
            rex.bark(Ellipsis)
        with pytest.raises(invocation.InvocationError):
            rex.bark(args)
        with pytest.raises(invocation.InvocationError):
            rex.bark(*args)
        with pytest.raises(invocation.InvocationError):
            rex.bark(kwargs)

        with pytest.raises(TypeError):
            rex.bark(1, 2)
        with pytest.raises(TypeError):
            rex.bark(wuff=1)
        with pytest.raises(TypeError):
            rex.bark(**kwargs)

    def testExpectingStarKwargs(self):
        rex = Dog()
        when(rex).bark(**kwargs).thenReturn('wuff')

        assert rex.bark(sound='miau') == 'wuff'

        with pytest.raises(invocation.InvocationError):
            rex.bark('miau')
        with pytest.raises(invocation.InvocationError):
            rex.bark(Ellipsis)
        with pytest.raises(invocation.InvocationError):
            rex.bark(kwargs)
        with pytest.raises(invocation.InvocationError):
            rex.bark(args)

        with pytest.raises(TypeError):
            rex.bark(wuff='miau')
        with pytest.raises(TypeError):
            rex.bark(**kwargs)

    def testExpectingEllipsis(self):
        rex = Dog()
        when(rex).bark(Ellipsis).thenReturn('wuff')

        assert rex.bark('miau') == 'wuff'
        with pytest.raises(TypeError):
            rex.bark('miau', 'miau')

        assert rex.bark(sound='miau') == 'wuff'
        with pytest.raises(TypeError):
            rex.bark(wuff='miau')

        assert rex.bark(Ellipsis) == 'wuff'
        assert rex.bark(args) == 'wuff'
        assert rex.bark(*args) == 'wuff'
        assert rex.bark(kwargs) == 'wuff'

        with pytest.raises(TypeError):
            rex.bark(**kwargs) == 'wuff'

    def testExpectingStarArgs(self):
        rex = Dog()
        when(rex).bark(*args).thenReturn('wuff')

        assert rex.bark('miau') == 'wuff'

        with pytest.raises(invocation.InvocationError):
            rex.bark(sound='miau')
        with pytest.raises(TypeError):
            rex.bark(wuff='miau')

        assert rex.bark(*args) == 'wuff'
        assert rex.bark(Ellipsis) == 'wuff'

        with pytest.raises(TypeError):
            rex.bark(**kwargs)


class TestEllipsises:

    # In python3 `bark(...)` is actually valid, but the tests must
    # be downwards compatible to python 2

    @pytest.mark.parametrize('call', [
        sig(),
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
        sig('Wuff', then='Wuff'),
        sig(then='Wuff'),
    ])
    def testEllipsisAsSoleArgumentAlwaysPasses(self, call):
        rex = mock()
        when(rex).bark(Ellipsis).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
        sig('Wuff', then='Wuff'),
    ])
    def testEllipsisAsSecondArgumentPasses(self, call):
        rex = mock()
        when(rex).bark('Wuff', Ellipsis).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig(),
        sig(then='Wuff'),
    ])
    def testEllipsisAsSecondArgumentRejections(self, call):
        rex = mock()
        when(rex).bark('Wuff', Ellipsis).thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig(),
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
    ])
    def testArgsAsSoleArgumentPasses(self, call):
        rex = mock()
        when(rex).bark(*args).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig('Wuff', then='Wuff'),
        sig(then='Wuff'),
    ])
    def testArgsAsSoleArgumentRejections(self, call):
        rex = mock()
        when(rex).bark(*args).thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
    ])
    def testArgsAsSecondArgumentPasses(self, call):
        rex = mock()
        when(rex).bark('Wuff', *args).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig(),
        sig('Wuff', then='Wuff'),
        sig(then='Wuff'),
    ])
    def testArgsAsSecondArgumentRejections(self, call):
        rex = mock()
        when(rex).bark('Wuff', *args).thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig('Wuff', then='Wuff'),
        sig('Wuff', 'Wuff', then='Wuff'),

    ])
    def testArgsBeforeConcreteKwargPasses(self, call):
        rex = mock()
        when(rex).bark('Wuff', *args, then='Wuff').thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig(),
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
        sig(then='Wuff'),

    ])
    def testArgsBeforeConcreteKwargRejections(self, call):
        rex = mock()
        when(rex).bark('Wuff', *args, then='Wuff').thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig(),
        sig(then='Wuff'),
        sig(then='Wuff', later='Waff')
    ])
    def testKwargsAsSoleArgumentPasses(self, call):
        rex = mock()
        when(rex).bark(**kwargs).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
        sig('Wuff', then='Wuff'),
        sig('Wuff', 'Wuff', then='Wuff'),
    ])
    def testKwargsAsSoleArgumentRejections(self, call):
        rex = mock()
        when(rex).bark(**kwargs).thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig(then='Wuff'),
        sig(then='Wuff', later='Waff'),
        sig(later='Waff', then='Wuff'),
    ])
    def testKwargsAsSecondKwargPasses(self, call):
        rex = mock()
        when(rex).bark(then='Wuff', **kwargs).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig(),
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
        sig('Wuff', then='Wuff'),
        sig('Wuff', 'Wuff', then='Wuff'),
        sig(first='Wuff', later='Waff')
    ])
    def testKwargsAsSecondKwargRejections(self, call):
        rex = mock()
        when(rex).bark(then='Wuff', **kwargs).thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig('Wuff', then='Waff'),
        sig('Wuff', 'Wuff', then='Waff'),
        sig('Wuff', then='Waff', later='Woff'),
        sig('Wuff', first="Wiff", then='Waff', later='Woff'),
        sig('Wuff', 'Wuff', then='Waff', later="Woff"),
    ])
    def testCombinedArgsAndKwargsPasses(self, call):
        rex = mock()
        when(rex).bark('Wuff', *args, then='Waff', **kwargs).thenReturn('Miau')

        assert rex.bark(*call.args, **call.kwargs) == 'Miau'

    @pytest.mark.parametrize('call', [
        sig(),
        sig('Wuff'),
        sig('Wuff', 'Wuff'),
        sig(later='Woff'),
        sig('Wuff', later='Woff'),
    ])
    def testCombinedArgsAndKwargsRejections(self, call):
        rex = mock()
        when(rex).bark('Wuff', *args, then='Waff', **kwargs).thenReturn('Miau')

        with pytest.raises(AssertionError):
            assert rex.bark(*call.args, **call.kwargs) == 'Miau'


    @pytest.mark.parametrize('call', [
        sig(Ellipsis),
    ])
    def testEllipsisMustBeLastThing(self, call):
        rex = mock()
        when(rex).bark(*call.args, **call.kwargs).thenReturn('Miau')

    @pytest.mark.parametrize('call', [
        sig(Ellipsis, 'Wuff'),
        sig(Ellipsis, then='Wuff'),
        sig(Ellipsis, 'Wuff', then='Waff'),
    ])
    def testEllipsisMustBeLastThingRejections(self, call):
        rex = mock()
        with pytest.raises(TypeError):
            when(rex).bark(*call.args, **call.kwargs).thenReturn('Miau')


    def testArgsMustUsedAsStarArg(self):
        rex = mock()
        with pytest.raises(TypeError):
            when(rex).bark(args).thenReturn('Miau')

    def testKwargsMustBeUsedAsStarKwarg(self):
        rex = mock()
        with pytest.raises(TypeError):
            when(rex).bark(kwargs).thenReturn('Miau')

        with pytest.raises(TypeError):
            when(rex).bark(*kwargs).thenReturn('Miau')

    def testNiceFormattingForEllipsis(self):
        m = mock()
        m.strict = False
        inv = invocation.StubbedInvocation(m, 'bark', None)
        inv(Ellipsis)

        assert repr(inv) == 'bark(...)'

    def testNiceFormattingForArgs(self):
        m = mock()
        m.strict = False
        inv = invocation.StubbedInvocation(m, 'bark', None)
        inv(*args)

        assert repr(inv) == 'bark(*args)'

    def testNiceFormattingForKwargs(self):
        m = mock()
        m.strict = False
        inv = invocation.StubbedInvocation(m, 'bark', None)
        inv(**kwargs)

        assert repr(inv) == 'bark(**kwargs)'




# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import functools
import inspect
import operator
from collections import deque

from . import invocation, signature, utils
from .mock_registry import mock_registry

__all__ = ['mock']

__tracebackhide__ = operator.methodcaller(
    "errisinstance",
    invocation.InvocationError
)

MYPY = False
if MYPY:
    from typing import Deque, List, Union
    RealInvocation = Union[
        invocation.RememberedInvocation,
        invocation.RememberedProxyInvocation
    ]


class _Dummy(object):
    # We spell out `__call__` here for convenience. All other magic methods
    # must be configured before use, but we want `mock`s to be callable by
    # default.
    def __call__(self, *args, **kwargs):
        return self.__getattr__('__call__')(*args, **kwargs)  # type: ignore[attr-defined]  # noqa: E501


def remembered_invocation_builder(mock, method_name, *args, **kwargs):
    invoc = invocation.RememberedInvocation(mock, method_name)
    return invoc(*args, **kwargs)


class Mock(object):
    def __init__(self, mocked_obj, strict=True, spec=None):
        self.mocked_obj = mocked_obj
        self.strict = strict
        self.spec = spec

        self.invocations = []  # type: List[RealInvocation]
        self.stubbed_invocations = deque()  \
            # type: Deque[invocation.StubbedInvocation]

        self._original_methods = {}
        self._methods_to_unstub = {}
        self._signatures_store = {}

    def remember(self, invocation):
        self.invocations.append(invocation)

    def finish_stubbing(self, stubbed_invocation):
        self.stubbed_invocations.appendleft(stubbed_invocation)

    def clear_invocations(self):
        self.invocations = []

    def get_original_method(self, method_name):
        return self._original_methods.get(method_name, None)

    # STUBBING

    def _get_original_method_before_stub(self, method_name):
        """
        Looks up the original method on the `spec` object and returns it
        together with an indication of whether the method is found
        "directly" on the `spec` object.

        This is used to decide whether the method should be stored as an
        original_method and should therefore be replaced when unstubbing.
        """
        if self.spec is None:
            return None, False

        try:
            return self.spec.__dict__[method_name], True
        except (AttributeError, KeyError):
            # Classes with defined `__slots__` and then no `__dict__` are not
            # patchable but if we catch the `AttributeError` here, we get
            # the better error message for the user.
            return getattr(self.spec, method_name, None), False

    def set_method(self, method_name, new_method):
        setattr(self.mocked_obj, method_name, new_method)

    def replace_method(self, method_name, original_method):

        def new_mocked_method(*args, **kwargs):
            return remembered_invocation_builder(
                self, method_name, *args, **kwargs)

        new_mocked_method.__name__ = method_name
        if original_method:
            new_mocked_method.__doc__ = original_method.__doc__
            new_mocked_method.__wrapped__ = original_method  # type: ignore[attr-defined]  # noqa: E501
            try:
                new_mocked_method.__module__ = original_method.__module__
            except AttributeError:
                pass

        if inspect.ismethod(original_method):
            new_mocked_method = utils.newmethod(
                new_mocked_method, self.mocked_obj
            )

        if isinstance(original_method, staticmethod):
            new_mocked_method = staticmethod(new_mocked_method)  # type: ignore[assignment]  # noqa: E501
        elif isinstance(original_method, classmethod):
            new_mocked_method = classmethod(new_mocked_method)  # type: ignore[assignment]  # noqa: E501
        elif (
            inspect.isclass(self.mocked_obj)
            and inspect.isclass(original_method)  # TBC: Inner classes
        ):
            new_mocked_method = staticmethod(new_mocked_method)  # type: ignore[assignment]  # noqa: E501

        self.set_method(method_name, new_mocked_method)

    def stub(self, method_name):
        try:
            self._methods_to_unstub[method_name]
        except KeyError:
            (
                original_method,
                was_in_spec
            ) = self._get_original_method_before_stub(method_name)
            if was_in_spec:
                # This indicates the original method was found directly on
                # the spec object and should therefore be restored by unstub
                self._methods_to_unstub[method_name] = original_method
            else:
                self._methods_to_unstub[method_name] = None

            self._original_methods[method_name] = original_method
            self.replace_method(method_name, original_method)

    def forget_stubbed_invocation(self, invocation):
        assert invocation in self.stubbed_invocations

        if len(self.stubbed_invocations) == 1:
            mock_registry.unstub(self.mocked_obj)
            return

        self.stubbed_invocations.remove(invocation)

        if not any(
            inv.method_name == invocation.method_name
            for inv in self.stubbed_invocations
        ):
            original_method = self._methods_to_unstub.pop(
                invocation.method_name
            )
            self.restore_method(invocation.method_name, original_method)

    def restore_method(self, method_name, original_method):
        # If original_method is None, we *added* it to mocked_obj, so we
        # must delete it here.
        if original_method:
            self.set_method(method_name, original_method)
        else:
            delattr(self.mocked_obj, method_name)

    def unstub(self):
        while self._methods_to_unstub:
            method_name, original_method = self._methods_to_unstub.popitem()
            self.restore_method(method_name, original_method)
        self.stubbed_invocations = deque()
        self.invocations = []

    # SPECCING

    def has_method(self, method_name):
        if self.spec is None:
            return True

        return hasattr(self.spec, method_name)

    def get_signature(self, method_name):
        if self.spec is None:
            return None

        try:
            return self._signatures_store[method_name]
        except KeyError:
            sig = signature.get_signature(self.spec, method_name)
            self._signatures_store[method_name] = sig
            return sig

    def eat_self(self, method_name):
        """Returns if the method will have a prepended self/class arg on call
        """
        try:
            original_method = self._original_methods[method_name]
        except KeyError:
            return False
        else:
            # If original_method is None, we *added* it to mocked_obj
            # and thus, it will eat self iff mocked_obj is a class.
            return (
                inspect.ismethod(original_method)
                or (
                    inspect.isclass(self.mocked_obj)
                    and not isinstance(original_method, staticmethod)
                    and not inspect.isclass(original_method)
                )
            )


class _OMITTED(object):
    def __repr__(self):
        return 'OMITTED'


OMITTED = _OMITTED()

def mock(config_or_spec=None, spec=None, strict=OMITTED):
    """Create 'empty' objects ('Mocks').

    Will create an empty unconfigured object, that you can pass
    around. All interactions (method calls) will be recorded and can be
    verified using :func:`verify` et.al.

    A plain `mock()` will be not `strict`, and thus all methods regardless
    of the arguments will return ``None``.

    .. note:: Technically all attributes will return an internal interface.
        Because of that a simple ``if mock().foo:`` will surprisingly pass.

    If you set strict to ``True``: ``mock(strict=True)`` all unexpected
    interactions will raise an error instead.

    You configure a mock using :func:`when`, :func:`when2` or :func:`expect`.
    You can also very conveniently just pass in a dict here::

        response = mock({'text': 'ok', 'raise_for_status': lambda: None})

    You can also create an empty Mock which is specced against a given
    `spec`: ``mock(requests.Response)``. These mock are by default strict,
    thus they raise if you want to stub a method, the spec does not implement.
    Mockito will also match the function signature.

    You can pre-configure a specced mock as well::

        response = mock({'json': lambda: {'status': 'Ok'}},
                        spec=requests.Response)

    Mocks are by default callable. Configure the callable behavior using
    `when`::

        dummy = mock()
        when(dummy).__call__(1).thenReturn(2)

    All other magic methods must be configured this way or they will raise an
    AttributeError.


    See :func:`verify` to verify your interactions after usage.

    """

    if type(config_or_spec) is dict:
        config, spec = config_or_spec, spec
    else:
        config, spec = {}, spec or config_or_spec

    if strict is OMITTED:
        strict = False if spec is None else True


    class Dummy(_Dummy):
        if spec:
            __class__ = spec  # make isinstance work

        def __getattr__(self, method_name):
            if strict:
                __tracebackhide__ = operator.methodcaller(
                    "errisinstance", AttributeError
                )

                raise AttributeError(
                    "'Dummy' has no attribute %r configured" % method_name)
            return functools.partial(
                remembered_invocation_builder, theMock, method_name)

        def __repr__(self):
            name = 'Dummy'
            if spec:
                name += spec.__name__
            return "<%s id=%s>" % (name, id(self))


    # That's a tricky one: The object we will return is an *instance* of our
    # Dummy class, but the mock we register will point and patch the class.
    # T.i. so that magic methods (`__call__` etc.) can be configured.
    obj = Dummy()
    theMock = Mock(Dummy, strict=strict, spec=spec)

    for n, v in config.items():
        if inspect.isfunction(v):
            invocation.StubbedInvocation(theMock, n)(Ellipsis).thenAnswer(v)
        else:
            setattr(Dummy, n, v)

    mock_registry.register(obj, theMock)
    return obj



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import pytest

from .test_base import TestBase
from mockito import (
    mock, when, verify, forget_invocations, inorder, VerificationError,
    ArgumentError, verifyNoMoreInteractions, verifyZeroInteractions,
    verifyNoUnwantedInteractions, verifyStubbedInvocationsAreUsed,
    any)
from mockito.verification import never


class VerificationTestBase(TestBase):
    def __init__(self, verification_function, *args, **kwargs):
        self.verification_function = verification_function
        TestBase.__init__(self, *args, **kwargs)

    def setUp(self):
        self.mock = mock()

    def testVerifies(self):
        self.mock.foo()
        self.mock.someOtherMethod(1, "foo", "bar")

        self.verification_function(self.mock).foo()
        self.verification_function(self.mock).someOtherMethod(1, "foo", "bar")

    def testVerifiesWhenMethodIsUsingKeywordArguments(self):
        self.mock.foo()
        self.mock.someOtherMethod(1, fooarg="foo", bararg="bar")

        self.verification_function(self.mock).foo()
        self.verification_function(self.mock).someOtherMethod(1, bararg="bar",
                                                              fooarg="foo")

    def testVerifiesDetectsNamedArguments(self):
        self.mock.foo(fooarg="foo", bararg="bar")

        self.verification_function(self.mock).foo(bararg="bar", fooarg="foo")
        try:
            self.verification_function(self.mock).foo(bararg="foo",
                                                      fooarg="bar")
            self.fail()
        except VerificationError:
            pass

    def testKeywordArgumentsOrderIsUnimportant(self):
        self.mock.blub(
            line="blabla", runs="55", failures="1", errors="2")
        self.verification_function(self.mock).blub(
            runs="55", failures="1", errors="2", line="blabla")

    def testFailsVerification(self):
        self.mock.foo("boo")

        self.assertRaises(VerificationError,
                          self.verification_function(self.mock).foo, "not boo")

    def testVerifiesAnyTimes(self):
        self.mock = mock()
        self.mock.foo()

        self.verification_function(self.mock).foo()
        self.verification_function(self.mock).foo()
        self.verification_function(self.mock).foo()

    def testVerifiesMultipleCalls(self):
        self.mock = mock()
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()

        self.verification_function(self.mock, times=3).foo()

    def testVerifiesMultipleCallsWhenMethodUsedAsFunction(self):
        self.mock = mock()
        f = self.mock.foo
        f(1, 2)
        f('foobar')

        self.verification_function(self.mock).foo(1, 2)
        self.verification_function(self.mock).foo('foobar')

    def testFailsVerificationOfMultipleCalls(self):
        self.mock = mock()
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()

        self.assertRaises(VerificationError,
                          self.verification_function(self.mock, times=2).foo)

    def testVerifiesUsingAnyMatcher(self):
        self.mock.foo(1, "bar")

        self.verification_function(self.mock).foo(1, any())
        self.verification_function(self.mock).foo(any(), "bar")
        self.verification_function(self.mock).foo(any(), any())

    def testVerifiesUsingAnyIntMatcher(self):
        self.mock.foo(1, "bar")

        self.verification_function(self.mock).foo(any(int), "bar")

    def testFailsVerificationUsingAnyIntMatcher(self):
        self.mock.foo(1, "bar")

        self.assertRaises(VerificationError,
                          self.verification_function(self.mock).foo, 1,
                          any(int))
        self.assertRaises(VerificationError,
                          self.verification_function(self.mock).foo, any(int))

    def testNumberOfTimesDefinedDirectlyInVerify(self):
        self.mock.foo("bar")

        self.verification_function(self.mock, times=1).foo("bar")

    def testFailsWhenTimesIsLessThanZero(self):
        self.assertRaises(ArgumentError, self.verification_function, None, -1)

    def testVerifiesAtLeastTwoWhenMethodInvokedTwice(self):
        self.mock.foo()
        self.mock.foo()

        self.verification_function(self.mock, atleast=2).foo()

    def testVerifiesAtLeastTwoWhenMethodInvokedFourTimes(self):
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()

        self.verification_function(self.mock, atleast=2).foo()

    def testFailsWhenMethodInvokedOnceForAtLeastTwoVerification(self):
        self.mock.foo()
        self.assertRaises(VerificationError,
                          self.verification_function(self.mock, atleast=2).foo)

    def testVerifiesAtMostTwoWhenMethodInvokedTwice(self):
        self.mock.foo()
        self.mock.foo()

        self.verification_function(self.mock, atmost=2).foo()

    def testVerifiesAtMostTwoWhenMethodInvokedOnce(self):
        self.mock.foo()

        self.verification_function(self.mock, atmost=2).foo()

    def testFailsWhenMethodInvokedFourTimesForAtMostTwoVerification(self):
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()

        self.assertRaises(VerificationError,
                          self.verification_function(self.mock, atmost=2).foo)

    def testVerifiesBetween(self):
        self.mock.foo()
        self.mock.foo()

        self.verification_function(self.mock, between=[1, 2]).foo()
        self.verification_function(self.mock, between=[2, 3]).foo()
        self.verification_function(self.mock, between=[1, 5]).foo()
        self.verification_function(self.mock, between=[2, 2]).foo()

    def testFailsVerificationWithBetween(self):
        self.mock.foo()
        self.mock.foo()
        self.mock.foo()

        self.assertRaises(VerificationError,
                          self.verification_function(self.mock,
                                                     between=[1, 2]).foo)
        self.assertRaises(VerificationError,
                          self.verification_function(self.mock,
                                                     between=[4, 9]).foo)

    def testFailsAtMostAtLeastAndBetweenVerificationWithWrongArguments(self):
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atleast=0)
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atleast=-5)
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atmost=0)
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atmost=-5)
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, between=[5, 1])
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, between=[-1, 1])
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, between=(0, 1, 2))
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, between=0)
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atleast=5, atmost=5)
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atleast=5, between=[1, 2])
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atmost=5, between=[1, 2])
        self.assertRaises(ArgumentError, self.verification_function,
                          self.mock, atleast=5, atmost=5, between=[1, 2])

    def runTest(self):
        pass


class VerifyTest(VerificationTestBase):
    def __init__(self, *args, **kwargs):
        VerificationTestBase.__init__(self, verify, *args, **kwargs)

    def testVerifyNeverCalled(self):
        verify(self.mock, never).someMethod()

    def testVerifyNeverCalledRaisesError(self):
        self.mock.foo()
        self.assertRaises(VerificationError, verify(self.mock, never).foo)


class InorderVerifyTest(VerificationTestBase):
    def __init__(self, *args, **kwargs):
        VerificationTestBase.__init__(self, inorder.verify, *args, **kwargs)

    def setUp(self):
        self.mock = mock()

    def testPassesIfOneIteraction(self):
        self.mock.first()
        inorder.verify(self.mock).first()

    def testPassesIfMultipleInteractions(self):
        self.mock.first()
        self.mock.second()
        self.mock.third()

        inorder.verify(self.mock).first()
        inorder.verify(self.mock).second()
        inorder.verify(self.mock).third()

    def testFailsIfNoInteractions(self):
        self.assertRaises(VerificationError, inorder.verify(self.mock).first)

    def testFailsIfWrongOrderOfInteractions(self):
        self.mock.first()
        self.mock.second()

        self.assertRaises(VerificationError, inorder.verify(self.mock).second)

    def testErrorMessage(self):
        self.mock.second()
        self.mock.first()
        self.assertRaisesMessage(
            '\nWanted first() to be invoked,\ngot    second() instead.',
            inorder.verify(self.mock).first)


    def testPassesMixedVerifications(self):
        self.mock.first()
        self.mock.second()

        verify(self.mock).first()
        verify(self.mock).second()

        inorder.verify(self.mock).first()
        inorder.verify(self.mock).second()

    def testFailsMixedVerifications(self):
        self.mock.second()
        self.mock.first()

        # first - normal verifications, they should pass
        verify(self.mock).first()
        verify(self.mock).second()

        # but, inorder verification should fail
        self.assertRaises(VerificationError, inorder.verify(self.mock).first)


class VerifyNoMoreInteractionsTest(TestBase):
    def testVerifies(self):
        mockOne, mockTwo = mock(), mock()
        mockOne.foo()
        mockTwo.bar()

        verify(mockOne).foo()
        verify(mockTwo).bar()
        verifyNoMoreInteractions(mockOne, mockTwo)

    def testFails(self):
        theMock = mock()
        theMock.foo()
        self.assertRaises(VerificationError, verifyNoMoreInteractions, theMock)


class VerifyZeroInteractionsTest(TestBase):
    def testVerifies(self):
        theMock = mock()
        verifyZeroInteractions(theMock)
        theMock.foo()
        self.assertRaises(
            VerificationError, verifyNoMoreInteractions, theMock)


class ClearInvocationsTest(TestBase):
    def testClearsInvocations(self):
        theMock1 = mock()
        theMock2 = mock()
        theMock1.do_foo()
        theMock2.do_bar()

        self.assertRaises(VerificationError, verifyZeroInteractions, theMock1)
        self.assertRaises(VerificationError, verifyZeroInteractions, theMock2)

        forget_invocations(theMock1, theMock2)

        verifyZeroInteractions(theMock1)
        verifyZeroInteractions(theMock2)

    def testPreservesStubs(self):
        theMock = mock()
        when(theMock).do_foo().thenReturn('hello')
        self.assertEqual('hello', theMock.do_foo())

        forget_invocations(theMock)

        self.assertEqual('hello', theMock.do_foo())


class TestRaiseOnUnknownObjects:
    @pytest.mark.parametrize('verification_fn', [
        verify,
        verifyNoMoreInteractions,
        verifyZeroInteractions,
        verifyNoUnwantedInteractions,
        verifyStubbedInvocationsAreUsed
    ])
    def testVerifyShouldRaise(self, verification_fn):
        class Foo(object):
            pass

        with pytest.raises(ArgumentError) as exc:
            verification_fn(Foo)
        assert str(exc.value) == "obj '%s' is not registered" % Foo






import sys

import pytest
from mockito import mock, when
from mockito.invocation import AnswerError

from . import module
from .test_base import TestBase


class Dog:
    def __init__(self, huge=False):
        self.huge = huge

    def bark(self):
        if self.huge:
            return "woof"
        else:
            return "waf ! waf ! waf ! waf ! waf ! waf !"

    @classmethod
    def class_bark(cls):
        return cls.__name__ + " woof"

    @staticmethod
    def static_bark(arg):
        return str(arg) + " woof"


class CallOriginalImplementationTest(TestBase):

    def testClassMethod(self):
        when(Dog).class_bark().thenCallOriginalImplementation()

        self.assertEqual("Dog woof", Dog.class_bark())

    def testStaticMethod(self):
        when(Dog).static_bark("wif").thenCallOriginalImplementation()
        self.assertEqual("wif woof", Dog.static_bark("wif"))

    def testStaticMethodOnInstance(self):
        dog = Dog()
        when(Dog).static_bark("wif").thenCallOriginalImplementation()
        self.assertEqual("wif woof", dog.static_bark("wif"))

    def testMethod(self):
        when(Dog).bark().thenCallOriginalImplementation()

        assert Dog(huge=True).bark() == "woof"

    def testMethodOnInstance(self):
        dog = Dog(huge=True)
        when(dog).bark().thenCallOriginalImplementation()

        assert dog.bark() == "woof"

    def testFunction(self):
        when(module).one_arg(Ellipsis).thenCallOriginalImplementation()
        assert module.one_arg("woof") == "woof"

    def testChain(self):
        when(module).one_arg(Ellipsis) \
                    .thenReturn("wif") \
                    .thenCallOriginalImplementation() \
                    .thenReturn("waf")
        assert module.one_arg("woof") == "wif"
        assert module.one_arg("woof") == "woof"
        assert module.one_arg("woof") == "waf"

    def testDumbMockHasNoOriginalImplementations(self):
        dog = mock()
        answer_selector = when(dog).bark()
        with pytest.raises(AnswerError) as exc:
            answer_selector.thenCallOriginalImplementation()

        if sys.version_info >= (3, 0):
            class_str_value = "mockito.mocking.mock.<locals>.Dummy"
        else:
            class_str_value = "mockito.mocking.Dummy"
        assert str(exc.value) == (
            "'<class '%s'>' "
            "has no original implementation for 'bark'."
        ) % class_str_value

    def testSpeccedMockHasOriginalImplementations(self):
        dog = mock({"huge": True}, spec=Dog)
        when(dog).bark().thenCallOriginalImplementation()
        assert dog.bark() == "woof"



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from mockito import any, unstub, verify, when
from mockito.verification import VerificationError

from .test_base import TestBase


class Dog:
    @staticmethod
    def bark():
        return "woof"

    @staticmethod
    def barkHardly(*args):
        return "woof woof"


class Cat:
    @staticmethod
    def meow():
        return "miau"


class StaticMethodsTest(TestBase):

    def tearDown(self):
        unstub()

    def testUnstubs(self):
        when(Dog).bark().thenReturn("miau")
        unstub()
        self.assertEqual("woof", Dog.bark())

    # TODO decent test case please :) without testing irrelevant implementation
    # details
    def testUnstubShouldPreserveMethodType(self):
        when(Dog).bark().thenReturn("miau!")
        unstub()
        self.assertTrue(isinstance(Dog.__dict__.get("bark"), staticmethod))

    def testStubs(self):
        self.assertEqual("woof", Dog.bark())

        when(Dog).bark().thenReturn("miau")

        self.assertEqual("miau", Dog.bark())

    def testStubsConsecutiveCalls(self):
        when(Dog).bark().thenReturn(1).thenReturn(2)

        self.assertEqual(1, Dog.bark())
        self.assertEqual(2, Dog.bark())
        self.assertEqual(2, Dog.bark())

    def testStubsWithArgs(self):
        self.assertEqual("woof woof", Dog.barkHardly(1, 2))

        when(Dog).barkHardly(1, 2).thenReturn("miau")

        self.assertEqual("miau", Dog.barkHardly(1, 2))

    def testStubsWithArgsOnInstance(self):
        dog = Dog()
        self.assertEqual("woof woof", dog.barkHardly(1, 2))

        when(Dog).barkHardly(1, 2).thenReturn("miau")

        self.assertEqual("miau", dog.barkHardly(1, 2))

    def testStubsButDoesNotMachArguments(self):
        self.assertEqual("woof woof", Dog.barkHardly(1, "anything"))

        when(Dog, strict=False).barkHardly(1, 2).thenReturn("miau")

        self.assertEqual(None, Dog.barkHardly(1))

    def testStubsMultipleClasses(self):
        when(Dog).barkHardly(1, 2).thenReturn(1)
        when(Dog).bark().thenReturn(2)
        when(Cat).meow().thenReturn(3)

        self.assertEqual(1, Dog.barkHardly(1, 2))
        self.assertEqual(2, Dog.bark())
        self.assertEqual(3, Cat.meow())

        unstub()

        self.assertEqual("woof", Dog.bark())
        self.assertEqual("miau", Cat.meow())

    def testVerifiesSuccesfully(self):
        when(Dog).bark().thenReturn("boo")

        Dog.bark()

        verify(Dog).bark()

    def testVerifiesWithArguments(self):
        when(Dog).barkHardly(1, 2).thenReturn("boo")

        Dog.barkHardly(1, 2)

        verify(Dog).barkHardly(1, any())

    def testFailsVerification(self):
        when(Dog).bark().thenReturn("boo")

        Dog.bark()

        self.assertRaises(VerificationError, verify(Dog).barkHardly, (1, 2))

    def testFailsOnInvalidArguments(self):
        when(Dog).bark().thenReturn("boo")

        Dog.barkHardly(1, 2)

        self.assertRaises(VerificationError, verify(Dog).barkHardly, (1, 20))

    def testFailsOnNumberOfCalls(self):
        when(Dog).bark().thenReturn("boo")

        Dog.bark()

        self.assertRaises(VerificationError, verify(Dog, times=2).bark)

    def testStubsAndVerifies(self):
        when(Dog).bark().thenReturn("boo")

        self.assertEqual("boo", Dog.bark())

        verify(Dog).bark()

    def testStubsTwiceAndUnstubs(self):
        when(Dog).bark().thenReturn(1)
        when(Dog).bark().thenReturn(2)

        self.assertEqual(2, Dog.bark())

        unstub()

        self.assertEqual("woof", Dog.bark())

    def testDoesNotVerifyStubbedCalls(self):
        when(Dog).bark().thenReturn(1)

        verify(Dog, times=0).bark()





import pytest

from mockito import when, when2, expect, verify, patch, mock, spy2
from mockito.invocation import InvocationError

class Dog(object):
    def bark(self):
        pass


class Unhashable(object):
    def update(self, **kwargs):
        pass

    def __hash__(self):
        raise TypeError("I'm immutable")


@pytest.mark.usefixtures('unstub')
class TestUserExposedInterfaces:

    def testWhen(self):
        whening = when(Dog)
        assert whening.__dict__ == {}

    def testExpect(self):
        expecting = expect(Dog)
        assert expecting.__dict__ == {}

    def testVerify(self):
        dummy = mock()
        verifying = verify(dummy)
        assert verifying.__dict__ == {}


    def testEnsureUnhashableObjectCanBeMocked(self):
        obj = Unhashable()
        when(obj).update().thenReturn(None)


@pytest.mark.usefixtures('unstub')
class TestPassAroundStrictness:

    def testReconfigureStrictMock(self):
        when(Dog).bark()  # important first call, inits theMock

        when(Dog, strict=False).waggle().thenReturn('Sure')
        expect(Dog, strict=False).weggle().thenReturn('Sure')


        with pytest.raises(InvocationError):
            when(Dog).wuggle()

        with pytest.raises(InvocationError):
            when(Dog).woggle()

        rex = Dog()
        assert rex.waggle() == 'Sure'
        assert rex.weggle() == 'Sure'

        # For documentation; the initial strict value of the mock will be used
        # here. So the above when(..., strict=False) just assures we can
        # actually *add* an attribute to the mocked object
        with pytest.raises(InvocationError):
            rex.waggle(1)

        verify(Dog).waggle()
        verify(Dog).weggle()

    def testReconfigureLooseMock(self):
        when(Dog, strict=False).bark()  # important first call, inits theMock

        when(Dog, strict=False).waggle().thenReturn('Sure')
        expect(Dog, strict=False).weggle().thenReturn('Sure')

        with pytest.raises(InvocationError):
            when(Dog).wuggle()

        with pytest.raises(InvocationError):
            when(Dog).woggle()

        rex = Dog()
        assert rex.waggle() == 'Sure'
        assert rex.weggle() == 'Sure'

        # For documentation; see test above. strict is inherited from the
        # initial mock. So we return `None`
        assert rex.waggle(1) is None

        verify(Dog).waggle()
        verify(Dog).weggle()

    # Where to put this test?
    def testEnsureAddedAttributesGetRemovedOnUnstub(self):
        with when(Dog, strict=False).wggle():
            pass

        with pytest.raises(AttributeError):
            getattr(Dog, 'wggle')


@pytest.mark.usefixtures('unstub')
class TestDottedPaths:

    def testWhen(self):
        when('os.path').exists('/Foo').thenReturn(True)

        import os.path
        assert os.path.exists('/Foo')

    def testWhen2(self):
        when2('os.path.exists', '/Foo').thenReturn(True)

        import os.path
        assert os.path.exists('/Foo')

    def testPatch(self):
        dummy = mock()
        patch('os.path.exists', dummy)

        import os.path
        assert os.path.exists('/Foo') is None

        verify(dummy).__call__('/Foo')

    def testVerify(self):
        when('os.path').exists('/Foo').thenReturn(True)

        import os.path
        os.path.exists('/Foo')

        verify('os.path', times=1).exists('/Foo')

    def testSpy2(self):
        spy2('os.path.exists')

        import os.path
        assert not os.path.exists('/Foo')

        verify('os.path', times=1).exists('/Foo')



# Copyright (c) 2008-2016 Szczepan Faber, Serhiy Oplakanets, Herr Kaste
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import operator

__all__ = ['never', 'VerificationError']


class VerificationError(AssertionError):
    '''Indicates error during verification of invocations.

    Raised if verification fails. Error message contains the cause.
    '''
    pass


__tracebackhide__ = operator.methodcaller("errisinstance", VerificationError)


class AtLeast(object):
    def __init__(self, wanted_count):
        self.wanted_count = wanted_count

    def verify(self, invocation, actual_count):
        if actual_count < self.wanted_count:
            raise VerificationError("\nWanted at least: %i, actual times: %i"
                                    % (self.wanted_count, actual_count))

    def __repr__(self):
        return "<%s wanted=%s>" % (type(self).__name__, self.wanted_count)

class AtMost(object):
    def __init__(self, wanted_count):
        self.wanted_count = wanted_count

    def verify(self, invocation, actual_count):
        if actual_count > self.wanted_count:
            raise VerificationError("\nWanted at most: %i, actual times: %i"
                                    % (self.wanted_count, actual_count))

    def __repr__(self):
        return "<%s wanted=%s>" % (type(self).__name__, self.wanted_count)

class Between(object):
    def __init__(self, wanted_from, wanted_to):
        self.wanted_from = wanted_from
        self.wanted_to = wanted_to

    def verify(self, invocation, actual_count):
        if actual_count < self.wanted_from or actual_count > self.wanted_to:
            raise VerificationError(
                "\nWanted between: [%i, %i], actual times: %i"
                % (self.wanted_from, self.wanted_to, actual_count))

    def __repr__(self):
        return "<%s [%s, %s]>" % (
            type(self).__name__, self.wanted_from, self.wanted_to)

class Times(object):
    def __init__(self, wanted_count):
        self.wanted_count = wanted_count

    def verify(self, invocation, actual_count):
        if actual_count == self.wanted_count:
            return

        if actual_count == 0:
            invocations = (
                [
                    invoc
                    for invoc in invocation.mock.invocations
                    if invoc.method_name == invocation.method_name
                ]
                or invocation.mock.invocations
                or ['Nothing']
            )
            raise VerificationError(
                """
Wanted but not invoked:

    %s

Instead got:

    %s

"""
                % (
                    invocation,
                    "\n    ".join(
                        str(invoc) for invoc in invocations
                    )
                )
            )
        else:
            if self.wanted_count == 0:
                raise VerificationError(
                    "\nUnwanted invocation of %s, times: %i"
                    % (invocation, actual_count))
            else:
                raise VerificationError("\nWanted times: %i, actual times: %i"
                                        % (self.wanted_count, actual_count))

    def __repr__(self):
        return "<%s wanted=%s>" % (type(self).__name__, self.wanted_count)

class InOrder(object):
    '''Verifies invocations in order.

    Verifies if invocation was in expected order, and if yes -- degrades to
    original Verifier (AtLeast, Times, Between, ...).
    '''

    def __init__(self, original_verification):
        '''

        @param original_verification: Original verification to degrade to if
                                      order of invocation was ok.
        '''
        self.original_verification = original_verification

    def verify(self, wanted_invocation, count):
        for invocation in wanted_invocation.mock.invocations:
            if not invocation.verified_inorder:
                if not wanted_invocation.matches(invocation):
                    raise VerificationError(
                        '\nWanted %s to be invoked,'
                        '\ngot    %s instead.' %
                        (wanted_invocation, invocation))
                invocation.verified_inorder = True
                break
        # proceed with original verification
        self.original_verification.verify(wanted_invocation, count)


never = 0



import mockito
from mockito import when, patch
import pytest

import numpy as np
from . import module


pytestmark = pytest.mark.usefixtures("unstub")


def xcompare(a, b):
    if isinstance(a, mockito.matchers.Matcher):
        return a.matches(b)

    return np.array_equal(a, b)


class TestEnsureNumpyWorks:
    def testEnsureNumpyArrayAllowedWhenStubbing(self):
        array = np.array([1, 2, 3])
        when(module).one_arg(array).thenReturn('yep')

        with patch(mockito.invocation.MatchingInvocation.compare, xcompare):
            assert module.one_arg(array) == 'yep'

    def testEnsureNumpyArrayAllowedWhenCalling(self):
        array = np.array([1, 2, 3])
        when(module).one_arg(Ellipsis).thenReturn('yep')
        assert module.one_arg(array) == 'yep'

